module Emitter where

import Parser
import Table
import Instruction

data Environment = Environment (Table, Int, Int) deriving Show

data EmitValue =
    EmitTemporary Int |
    EmitVariable Int String |
    EmitConstant Int |
    EmitLabel Int |
    EmitRegisterA |
    EmitNone

emitError = error . ("emitter error: " ++)

envTable (Environment (table, offset, label)) = table
envOffset (Environment (table, offset, label)) = offset
envLabel (Environment (table, offset, label)) = label
incEnvOffset (Environment (table, offset, label)) = 
    if offset < 0 then
        emitError "memory over"
    else
        (Environment (table, offset - 1, label))
decEnvOffset (Environment (table, offset, label)) = 
    (Environment (table, offset + 1, label))
incEnvLabel (Environment (table, offset, label)) = (Environment (table, offset , label + 1))

emitValueOfTriple (ev, code, env) = ev
codeOfTriple (ev, code, env) = code
envOfTriple (ev, code, env) = env

emit tree = emitProgram tree (Environment (initialTable "undeclared variable ", 0xf, 0))
--emit tree = resolvFlagReset $ emitProgram tree (Environment (initialTable "undeclared variable ", 0xf, 0))

emitProgram (Program functions) env = emitConcat emitFunction env functions
emitConcat f env l = concat $ map (\x -> f x env) l

emitFunction (Function (name, ftype, arguments, block)) env =
    codeOfTriple (emitBlock block env)

emitBlock (Block (declarations, statements)) env =
    (emitValueOfTriple statementsResult,
     codeOfTriple statementsResult,
     Environment (envTable env, envOffset env, envLabel (envOfTriple statementsResult)))
    where
      newEnv = emitDeclarations declarations env
      statementsResult = emitStatements statements newEnv

emitStatements = emitSequence emitStatement

emitSequence ef (t:ts) env = 
    (if ts == [] then EmitNone else emitValueOfTriple emitResult,
     (codeOfTriple emitResult) ++ codeOfTriple nextEmitResult,
     envOfTriple nextEmitResult)
    where
      emitResult = ef t env
      nextEmitResult = emitSequence ef ts (envOfTriple emitResult)
emitSequence ef [] env = (EmitNone, [], env)

emitStatement tree env = 
    case tree of
      (Substitution ((Variable variable), expression)) ->
          let 
              emitResult = emitExpression expression env
              emitVal = emitValueOfTriple emitResult
              code = codeOfTriple emitResult
              variableOffset = offsetTable (envTable env) variable
          in
            (EmitNone,
            code ++ 
            toRegisterA emitVal ++
            [(OperandInt1 "TIY" variableOffset, variable ++ " = ..."),
             (Operand0 "AM", "")],
             env)
      (If (expression, statements)) ->
          let
              condResult = emitExpression expression env
              condVal = emitValueOfTriple condResult
              condCode = codeOfTriple condResult
              thenResult = emitStatement statements (envOfTriple condResult)
              newEnv = envOfTriple thenResult
              labelNo = envLabel newEnv
              label1 = emitLabel labelNo
          in
            (EmitNone,
            condCode ++ toRegisterA condVal ++
            [(OperandInt1 "AIA" 0xf, "if judege"),
             (OperandInt1 "AIA" 1, ""),
             (OperandString1 "JUMP" label1, "")] ++ 

             codeOfTriple thenResult ++

            [(Label label1, "if end")],
             incEnvLabel newEnv)

      (IfElse (expression, thenStatements, elseStatements)) ->
          let
              condResult = emitExpression expression env
              condVal = emitValueOfTriple condResult
              condCode = codeOfTriple condResult
              thenResult = emitStatement thenStatements (envOfTriple condResult)
              elseResult = emitStatement elseStatements (envOfTriple thenResult)
              newEnv = envOfTriple elseResult
              labelNo = envLabel newEnv
              label1 = emitLabel labelNo
              label2 = emitLabel (labelNo + 1)
          in
            (EmitNone,
            condCode ++ toRegisterA condVal ++
            [(OperandInt1 "CIA" 0, "if else judge"),
             (OperandString1 "JUMP" label1, "")] ++ 
             codeOfTriple elseResult ++
            [(FlagResetDontCare, "")] ++
            [(OperandString1 "JUMP" label2, "if else jump"),
             (Label label1, "")] ++

             codeOfTriple thenResult ++

            [(Label label2, "if else end")],
             (incEnvLabel . incEnvLabel) newEnv)

      (While (expression, statements)) ->
          let
              condResult = emitExpression expression env
              condVal = emitValueOfTriple condResult
              condCode = codeOfTriple condResult
              thenResult = emitStatement statements (envOfTriple condResult)
              newEnv = envOfTriple thenResult
              labelNo = envLabel newEnv
              label1 = emitLabel labelNo
              label2 = emitLabel (labelNo + 1)
          in
            (EmitNone,
            [(Label label1, "while start")] ++
            condCode ++ toRegisterA condVal ++

            [(OperandInt1 "AIA" 0xf, "while judege"),
             (OperandInt1 "AIA" 1, ""),
             (OperandString1 "JUMP" label2, "")] ++
                                               
            codeOfTriple thenResult ++

            [(FlagResetDontCare, "")] ++
            [(OperandString1 "JUMP" label1, "while loop jump"),
             (Label label2, "while end")],
             (incEnvLabel . incEnvLabel) newEnv)


      (StatementCall0 "exit") ->
          let
              labelNo = envLabel env
              label1 = emitLabel labelNo
          in
            (EmitNone,
             [(FlagResetDontCare, ""),
              (Label label1, ""),
             (OperandString1 "JUMP" label1, "exit loop")],
             incEnvLabel env)

      (StatementCall0 fname) ->
          (EmitNone, [(FlagResetDontCare, "")] ++ emitStatementCall0 fname, env)

      (StatementCall1 (fname, expression)) ->
          let
              emitResult = emitExpression expression env
          in
            (EmitNone,
             codeOfTriple emitResult ++
             toRegisterA (emitValueOfTriple emitResult) ++
             emitStatementCall1 fname,
             envOfTriple emitResult)
              
      (StatementBlock block) -> emitBlock block env

      _ -> emitError ("unkown statement")

emitStatementCall0 fn@"resetNumberLed" = [(OperandString1 "CAL" "RSTO", "call " ++ fn ++ "()")]
emitStatementCall0 fn@"beepEndSound" = [(OperandString1 "CAL" "ENDS", "call " ++ fn ++ "()")]
emitStatementCall0 fn@"beepErrorSound" = [(OperandString1 "CAL" "ERRS", "call " ++ fn ++ "()")]
emitStatementCall0 fn@"beepShortSound" = [(OperandString1 "CAL" "SHTS", "call " ++ fn ++ "()")]
emitStatementCall0 fn@"beepLongSound" = [(OperandString1 "CAL" "LONS", "call " ++ fn ++ "()")]
emitStatementCall0 fname = emitError ("unkown function " ++ fname ++ "(void)")

emitStatementCall1 fn@"setNumberLed" = [(FlagResetRegistersSafe, "")] ++ [(Operand0 "AO", "call " ++ fn ++ "()")]
emitStatementCall1 fn@"setLed" = [(Operand0 "CY", ""), (OperandString1 "CAL" "SETR", "call " ++ fn ++ "()")]
emitStatementCall1 fn@"resetLed" = [(Operand0 "CY", ""), (OperandString1 "CAL" "RSTR", "call " ++ fn ++ "()")]
emitStatementCall1 fn@"beep" = [(FlagResetRegistersSafe, "")] ++ [(OperandString1 "CAL" "SUND", "call " ++ fn ++ "()")]
emitStatementCall1 fn@"sleep" = [(FlagResetRegistersSafe, "")] ++ [(OperandString1 "CAL" "TIMR", "call " ++ fn ++ "()")]
emitStatementCall1 fname = emitError ("unkown function " ++ fname ++ "(int)")

emitExpression tree env = 
    case tree of
      (Number i) -> (EmitConstant i, [], env)
      (VariableValue (Variable s)) -> (EmitVariable (offsetTable (envTable env) s) s, [], env)
      Operation2 (operator , expression1, expression2) ->
          let
              emitResult1 = emitExpression expression1 env
              emitResult2 = emitExpression expression2 (incEnvOffset (envOfTriple emitResult1))
              newEnv = decEnvOffset (envOfTriple emitResult2)
              offset = envOffset newEnv

              reverseEmitResult1 = emitExpression expression2 env
              reverseEmitResult2 = emitExpression expression1 (incEnvOffset (envOfTriple reverseEmitResult1))
              reverseNewEnv = decEnvOffset (envOfTriple reverseEmitResult2)
              reverseOffset = envOffset reverseNewEnv
          in
            case operator of
              o | any (o==) ["+", "-", "!="] -> let op = if o == "!=" then "-" else o in
                  (EmitRegisterA,
                   codeOfTriple emitResult1 ++
                  (toRegisterA (emitValueOfTriple emitResult1)) ++
                  [(OperandInt1 "TIY" offset, ""),
                   (Operand0 "AM", "")] ++
                   codeOfTriple emitResult2 ++
                  (toRegisterA (emitValueOfTriple emitResult2)) ++
                  [(OperandInt1 "TIY" offset, ""),
                   (Operand0 ("M" ++ op), "Operator '" ++ operator ++ "'")],
                   newEnv)


              o | any (o==) [">=", "<=", ">", "<"] ->
                  let
                      emitResultOp1 = (if o == ">=" || o == "<" then emitResult1 else reverseEmitResult1)
                      emitResultOp2 = (if o == ">=" || o == "<" then emitResult2 else reverseEmitResult2)
                      targetNewEnv = (if o == ">=" || o == "<" then newEnv else reverseNewEnv)
                                      
                      labelNo = envLabel targetNewEnv
                      label1 = emitLabel labelNo
                  in
                    (EmitRegisterA,
                     codeOfTriple emitResultOp1 ++
                     (toRegisterA (emitValueOfTriple emitResultOp1)) ++
                     [(OperandInt1 "TIY" offset, ""),
                      (Operand0 "AM", "")] ++
                     codeOfTriple emitResultOp2 ++
                     (toRegisterA (emitValueOfTriple emitResultOp2)) ++
                     [(OperandInt1 "TIY" offset, ""),
                      (Operand0 "CH", ""),
                      (if o == "<=" || o == ">="
                       then (OperandInt1 "TIA" 0, "set false")
                       else (OperandInt1 "TIA" 1, "set true")
                      ),
                      (Operand0 "CH", ""),
                      (Operand0 ("M-"), "Operator '" ++ operator ++ "'"),
                      (OperandString1 "JUMP" label1, ""),
                      (if o == "<=" || o == ">="
                       then (OperandInt1 "TIA" 1, "set true")
                       else (OperandInt1 "TIA" 0, "set false")
                      ),
                      (Operand0 "CH", ""),
                      (Label label1, ""),
                      (Operand0 "CH", "")],
                     incEnvLabel targetNewEnv)

              "==" ->
                  (EmitRegisterA,
                   codeOfTriple emitResult1 ++
                   (toRegisterA (emitValueOfTriple emitResult1)) ++
                   [(OperandInt1 "TIY" offset, ""),
                    (Operand0 "AM", "")] ++
                   codeOfTriple emitResult2 ++
                   (toRegisterA (emitValueOfTriple emitResult2)) ++
                   [(OperandInt1 "TIY" offset, ""),
                    (OperandString1 "CAL" "CHNG", "set false Ar'"),
                    (OperandInt1 "TIA" 0, ""),
                    (OperandInt1 "TIY" offset, ""),
                    (OperandString1 "CAL" "CHNG", ""),
                    (Operand0 ("M-"), "Operator '" ++ operator ++ "'"),
                    (OperandInt1 "AIA" 0xf, ""),
                    (OperandString1 "CAL" "CHNG", "")],
                   newEnv)

              _ -> emitError ("operator " ++ operator ++ " not supported")

      (InputKeyCall vname) ->
          let
              labelNo = envLabel env
              label1 = emitLabel labelNo
              offset = envOffset env
              variableOffset = offsetTable (envTable env) vname
          in
            (EmitRegisterA,
             [(OperandInt1 "TIA" 0, "inputKey start"),
              (Operand0 "CH" , ""),
              (Operand0 "KA", ""),
              (OperandString1 "JUMP" label1, ""),
              (OperandInt1 "TIY" variableOffset, ""),
              (Operand0 "AM", ""),
              (OperandInt1 "TIA" 1, ""),
              (Operand0 "CH" , ""),
              (Label label1, ""),
              (Operand0 "CH" , "")],
             incEnvLabel env)


emitDeclarations declarations env = foldl (\x y -> emitDeclaration y x) env declarations

emitDeclaration (Declaration (dtype, variables)) env =
    foldl (\x y -> emitDeclarationVariable dtype y x) env variables

emitDeclarationVariable TypeInt (Variable vname) env =
    if offset < 0 then
        emitError "memory over"
    else
        Environment (newTable, offset - 1, envLabel env)
    where
      offset = envOffset env
      newTable = updateTable (envTable env) vname offset

toRegisterA EmitRegisterA = []
toRegisterA (EmitTemporary i) =
    [(OperandInt1 "TIY" i, "Temporary offset " ++ show i) , (Operand0 "MA", "")]
toRegisterA (EmitVariable i s) =
    [(OperandInt1 "TIY" i, "Variable '" ++ s ++ "' (offset " ++ show i ++ ")") , (Operand0 "MA", "")]
toRegisterA (EmitConstant i) = 
    [(OperandInt1 "TIA" i, "Constant " ++ show i)]

emitLabel = ("LABEL" ++) . show
